import os
import pygame

from .widgets import StatusBar
from .colors import GREY
from .buttons import BigSmallBut, SaveBut, CopyBut, BackBut, ClearBut
from . import shapes


class Snap:
    level = int(os.environ.get("SNAPPING", 8))

    @classmethod
    def getSnapped(cls, coord: tuple[int, int]) -> list[int]:
        lvl = cls.level
        if not lvl:
            return list(coord)
        return [((lvl // 2 + coord[0]) // lvl) * lvl, ((lvl // 2 + coord[1]) // lvl) * lvl]


class GUI:
    statusbar_height = 40

    def __init__(self, background: pygame.Surface) -> None:
        self.objects: list[shapes.Shape] = []
        self.dragging = False
        self.but_undo = BackBut(self)
        self.but_save = SaveBut(self)
        self.but_clear = ClearBut(self)
        self.but_copy = CopyBut(self)
        self.dirty_statusbar = True
        self.dirty_annotation = True
        buttons = [
            self.but_undo,
            self.but_copy,
            self.but_save,
            BigSmallBut(self),
            self.but_clear,
        ]
        bg_rect = background.get_rect()
        self.statusbar_height = StatusBar.getHeight(bg_rect.width, buttons)
        self.screen = pygame.display.set_mode(
            (
                max(
                    bg_rect.width,
                    StatusBar.estimateWidth(buttons, self.statusbar_height),
                ),
                bg_rect.height + self.statusbar_height,
            ),
            pygame.DOUBLEBUF,
            vsync=1,
        )

        self.background = background
        self.statusbar_surface = pygame.Surface((self.screen.get_width(), self.statusbar_height), pygame.SRCALPHA).convert_alpha()
        self.statusbar = StatusBar(self.statusbar_surface, buttons, self.statusbar_height)
        self.annotation_overlay = pygame.Surface(self.background.get_size(), pygame.SRCALPHA).convert_alpha()

    def get_annotated_image(self):
        surface = pygame.Surface(self.background.get_size(), pygame.SRCALPHA)
        surface.blit(self.background, (0, 0))
        surface.blit(self.annotation_overlay, (0, 0))
        return surface

    def handle_event(self, event):
        if event.pos[1] < self.statusbar_height:
            if self.statusbar.handle_event(event):
                self.dirty_statusbar = True
        else:
            if event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 1:  # left mouse button
                    self.dragging = True

                    # Add a shape to the list of shapes
                    start_pos = Snap.getSnapped(event.pos)
                    start_pos[1] -= self.statusbar_height
                    shape = self.statusbar.selected_shape(
                        color=self.statusbar.selected_color,
                        start=start_pos,
                        end=start_pos,
                    )
                    self.objects.append(shape)
                    self.dirty_annotation = True
            elif event.type == pygame.MOUSEBUTTONUP:
                if event.button == 1:  # left mouse button
                    if self.dragging:
                        pos = Snap.getSnapped(event.pos)
                        pos[1] -= self.statusbar_height
                        self.objects[-1].end = pos
                        self.dragging = False
                        self.dirty_annotation = True
            elif event.type == pygame.MOUSEMOTION:
                if self.dragging:
                    pos = list(event.pos)
                    pos[1] -= self.statusbar_height
                    if self.objects:
                        self.objects[-1].end = pos
                        self.dirty_annotation = True

    def draw(self, force=False):
        dirty = False
        if force or self.dirty_statusbar:
            dirty = True
            self.statusbar_surface.fill(GREY)
            self.statusbar.draw()
            self.screen.blit(self.statusbar_surface, (0, 0))
            self.dirty_statusbar = False

        if force or self.dirty_annotation:
            dirty = True
            self.screen.blit(self.background, (0, self.statusbar_height))
            self.annotation_overlay.fill((0, 0, 0, 0))

            for shape in self.objects:
                shape.draw(self.annotation_overlay)

            self.screen.blit(self.annotation_overlay, (0, self.statusbar_height))
            self.dirty_annotation = False

        if dirty:
            pygame.display.flip()


def main(image_path: str):
    pygame.init()
    pygame.display.set_caption("Draw Shapes")

    background = pygame.image.load(image_path)

    # Create the GUI
    gui = GUI(background)

    clock = pygame.time.Clock()
    # Game loop
    running = True

    while running:
        # Update the screen
        gui.draw()
        # Handle events
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.WINDOWMOVED:  # FIXME: workaround for black screen
                gui.draw(force=True)
            else:
                if event.type in (
                    pygame.MOUSEBUTTONDOWN,
                    pygame.MOUSEBUTTONUP,
                    pygame.MOUSEMOTION,
                    pygame.BUTTON_WHEELDOWN,
                    pygame.BUTTON_WHEELUP,
                ):
                    gui.handle_event(event)
                else:
                    if event.type == pygame.KEYDOWN:
                        if event.key == pygame.K_s:
                            gui.but_save.execute()
                        elif event.key == pygame.K_ESCAPE:
                            running = False
                        elif event.key == pygame.K_c:
                            gui.but_clear.execute()
                        elif event.key == pygame.K_BACKSPACE:
                            gui.but_undo.execute()
                        elif event.key == pygame.K_r:
                            gui.statusbar.selected_shape = shapes.Rectangle
                            gui.dirty_statusbar = True
                        elif event.key == pygame.K_a:
                            gui.statusbar.selected_shape = shapes.Arrow
                            gui.dirty_statusbar = True
                        elif event.key == pygame.K_e:
                            gui.statusbar.selected_shape = shapes.Bullet
                            gui.dirty_statusbar = True
        clock.tick(60)

    gui.but_copy.execute()
    # Quit pygame
    pygame.quit()



import math
import pygame

from .colors import BLACK, WHITE

WIDGET_SCALE = 1
SUPERSAMPLE = 4

OUTPUT_FILENAME = "/tmp/annotated.jpg"


def make_color_shape(color: tuple[int, int, int], width: int, height: int):
    w = width * SUPERSAMPLE
    h = height * SUPERSAMPLE
    surf = pygame.Surface((w, h), pygame.SRCALPHA)
    surf.fill(list(color) + [0])
    pygame.draw.circle(surf, (50, 50, 50), (w / 2 - 4, h / 2 + 4), int(0.4 * w))
    pygame.draw.circle(surf, color, (w / 2, h / 2), int(0.4 * w))
    return pygame.transform.smoothscale(surf, (width, height))


class Shape:
    _name = "unknown"
    _surface = None
    shadow = (3 * SUPERSAMPLE, 3 * SUPERSAMPLE)
    shadow_color = (0, 0, 0, 150)

    def __init__(
        self,
        color: tuple[int, int, int] | list[int],
        start: tuple[int, int] | list[int],
        end: tuple[int, int] | list[int],
        isDummy=False,
    ):
        self.color = color
        self.start = start
        self.end = end
        self.isDummy = isDummy

    @property
    def inv_color(self):
        r, g, b = self.color

        # Calculate relative luminance
        L = (0.2126 * r + 0.7152 * g + 0.0722 * b) / 255.0

        # Choose white or black text color based on luminance
        if L > 0.5:
            return BLACK
        else:
            return WHITE

    @property
    def rect(self):
        return pygame.Rect(
            self.start, (self.end[0] - self.start[0], self.end[1] - self.start[1])
        )

    def draw(self, surface: pygame.Surface):
        raise NotImplementedError(
            "Missing draw implementation for the {} class.".format(self._name)
        )

    def instance_removed(self):
        pass


class Arrow(Shape):
    thickness = 5
    arrowhead_size = 20
    _name = "arrow"
    _old_pos = None
    _ssurface = None

    @property
    def fixed_size(self):
        if self.isDummy:
            return self.arrowhead_size
        else:
            return self.arrowhead_size * WIDGET_SCALE

    def draw(self, surface):
        start = [SUPERSAMPLE * i for i in self.start]
        end = [SUPERSAMPLE * i for i in self.end]

        if not self._surface or self._old_pos != (start, end, WIDGET_SCALE):
            self._old_pos = (start, end, WIDGET_SCALE)

            if self.isDummy:
                # add some padding
                end[0] -= 5
                end[1] -= 5
                start[0] += 10
                start[1] += 10

            # Calculate the angle of the line from start to end
            angle = math.atan2(end[1] - start[1], end[0] - start[0])
            if not self._ssurface:
                supersampled_surface = pygame.Surface(
                    tuple(SUPERSAMPLE * i for i in surface.get_size()), pygame.SRCALPHA
                )
                self._ssurface = supersampled_surface
            else:
                self._ssurface.fill((0, 0, 0, 0))

            # Calculate the endpoint of the line from start to end
            line_length = math.sqrt((end[1] - start[1]) ** 2 + (end[0] - start[0]) ** 2)
            end_point = (
                int(start[0] + line_length * math.cos(angle)),
                int(start[1] + line_length * math.sin(angle)),
            )

            if self.isDummy:
                sz = int(self.fixed_size * SUPERSAMPLE * 0.7)
            else:
                sz = (
                    self.fixed_size
                    * SUPERSAMPLE
                    * (WIDGET_SCALE * 0.5 if WIDGET_SCALE > 1 else 1)
                )
            # shadow
            if not self.isDummy:
                pygame.draw.polygon(
                    self._ssurface,
                    self.shadow_color,
                    (
                        end_point,
                        (
                            (end_point[0] - sz * math.cos(angle - math.pi / 6))
                            - self.shadow[0],
                            (end_point[1] - sz * math.sin(angle - math.pi / 6))
                            + self.shadow[1],
                        ),
                        (
                            (end_point[0] - sz * math.cos(angle + math.pi / 6))
                            - self.shadow[0],
                            (end_point[1] - sz * math.sin(angle + math.pi / 6))
                            + self.shadow[1],
                        ),
                    ),
                )
                sline_length = line_length - (sz // 2)
                send_point = (
                    int(start[0] - self.shadow[0] + sline_length * math.cos(angle)),
                    int(start[1] + self.shadow[1] + sline_length * math.sin(angle)),
                )
                # Draw the line from start to end
                pygame.draw.line(
                    self._ssurface,
                    self.shadow_color,
                    start,
                    send_point,
                    self.thickness * WIDGET_SCALE * SUPERSAMPLE + 5,
                )
            # real shape
            # Draw a circle t point A
            pygame.draw.circle(
                self._ssurface,
                self.color,
                start,
                2 * (1 if self.isDummy else WIDGET_SCALE * SUPERSAMPLE),
            )
            # Draw the arrowhead at point B
            pygame.draw.polygon(
                self._ssurface,
                self.color,
                (
                    (end_point[0], end_point[1]),
                    (
                        end_point[0] - sz * math.cos(angle - math.pi / 6),
                        end_point[1] - sz * math.sin(angle - math.pi / 6),
                    ),
                    (
                        end_point[0] - sz * math.cos(angle + math.pi / 6),
                        end_point[1] - sz * math.sin(angle + math.pi / 6),
                    ),
                ),
            )
            line_length -= sz // 2
            end_point = (
                int(start[0] + line_length * math.cos(angle)),
                int(start[1] + line_length * math.sin(angle)),
            )
            # Draw the line from start to end
            pygame.draw.line(
                self._ssurface,
                self.color,
                start,
                end_point,
                self.thickness * (2 if self.isDummy else WIDGET_SCALE * SUPERSAMPLE),
            )
            self._surface = pygame.transform.smoothscale(
                self._ssurface, surface.get_size()
            )
        surface.blit(self._surface, (0, 0))


class Bullet(Shape):
    _name = "bullet"
    size = 12
    _counter = 0

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        if not self.isDummy:
            Bullet._counter += 1
        self.counter = Bullet._counter
        self._old_scale = 0
        self._surface = None

    def instance_removed(self):
        Bullet._counter -= 1

    @property
    def corrected_size(self):
        return self.size * SUPERSAMPLE

    def draw(self, surface):
        if WIDGET_SCALE != self._old_scale:
            self._font = pygame.font.SysFont(
                "Arial",
                48
                * (
                    1
                    if self.isDummy
                    else (WIDGET_SCALE // 2 if WIDGET_SCALE > 1 else 1)
                ),
                bold=True,
            )
            self.text = self._font.render(
                "2" if self.isDummy else str(self.counter), True, self.inv_color
            )
            self._surface = None

        if not self._surface:
            border_sz = int(
                SUPERSAMPLE * 3 * (WIDGET_SCALE * 0.5 if WIDGET_SCALE > 1 else 1)
            )
            side = max(self.text.get_size()[0] + 10, self.corrected_size) + border_sz

            supersampled_surface = pygame.Surface(
                (side + border_sz, side + border_sz), pygame.SRCALPHA
            )
            supersampled_surface.fill((0, 0, 0, 0))
            corrected_rect = pygame.Rect(0, 0, side, side)
            corrected_rect[0] += border_sz // 2
            corrected_rect[1] += border_sz // 2

            if not self.isDummy:  # draw shadow
                pygame.draw.circle(
                    supersampled_surface,
                    self.shadow_color,
                    (corrected_rect.center[0] - 4, corrected_rect.center[1] + 4),
                    (side // 2),
                )
            pygame.draw.circle(
                supersampled_surface, self.inv_color, corrected_rect.center, (side // 2)
            )
            pygame.draw.circle(
                supersampled_surface,
                self.color,
                corrected_rect.center,
                (side - border_sz) // 2,
            )

            pos = list(corrected_rect.center)
            text_size = self.text.get_size()
            pos[0] -= text_size[0] // 2
            pos[1] -= text_size[1] // 2
            supersampled_surface.blit(self.text, pos)
            sz = self.rect.size[0] if self.isDummy else (self.size * WIDGET_SCALE * 2)
            self._surface = pygame.transform.smoothscale(supersampled_surface, (sz, sz))

        if self.isDummy:
            surface.blit(
                self._surface, [i - self.rect.width // 2 for i in self.rect.center]
            )
        else:
            surface.blit(
                self._surface,
                (
                    self.start[0] - self.size * WIDGET_SCALE,
                    self.start[1] - self.size * WIDGET_SCALE,
                ),
            )


class Rectangle(Shape):
    _name = "rectangle"
    thickness = 3

    def draw(self, surface):
        if self.isDummy:
            rect = self.rect.copy()
            rect.width -= 6
            rect.height -= 6
            rect.x += 3
            rect.y += 3
            pygame.draw.rect(surface, self.color, rect, width=2)
        else:
            pygame.draw.rect(
                surface,
                list(self.color) + [50],
                self.rect,
            )
            pygame.draw.rect(
                surface,
                self.color,
                self.rect,
                width=1 if self.isDummy else self.thickness * WIDGET_SCALE,
            )

    @property
    def rect(self):
        x = min(self.start[0], self.end[0])
        y = min(self.start[1], self.end[1])
        width = max(self.start[0], self.end[0]) - x
        height = max(self.start[1], self.end[1]) - y
        return pygame.Rect(x, y, width, height)


all_shapes = [Rectangle, Arrow, Bullet]



#!/bin/env python
import os
import sys


def run():
    try:
        image = sys.argv[1]
    except IndexError:
        print("Usage: python %s <image>" % sys.argv[0])
    else:
        os.environ["SDL_VIDEODRIVER"] = "x11"
        from .gui import main

        main(image)


if __name__ == "__main__":
    run()



import pygame
import itertools

from . import shapes
from .colors import BLACK, GREY, SELECTED_COLOR


class Icon:
    _index = 0

    def __init__(self, x, y, size):
        self.rect = pygame.Rect(x, y, size, size)
        self.index = Icon._index
        Icon._index += 1
        self.shape = None


class Button:
    def __init__(self, gui, label="X"):
        self.gui = gui
        font = pygame.font.SysFont("Symbols Nerd Font", 20)
        self.text = font.render(label, True, BLACK)
        self.bg = GREY

    def draw(self, surface, rect):
        pygame.draw.rect(surface, self.bg, rect, border_radius=5)
        sz = self.text.get_size()
        pos = tuple((rect.center[0] - sz[0] // 2, rect.center[1] - sz[1] // 2))
        surface.blit(self.text, pos)


class StatusBar:
    margin = 5
    separation = 20
    # List of distinctive colors
    available_colors = [
        (128, 0, 0),
        (230, 25, 75),
        (255, 255, 25),
        (60, 180, 75),
        (70, 240, 240),
        (240, 50, 230),
        (0, 130, 200),
        (255, 255, 255),
        (0, 0, 0),
    ]
    available_shapes = shapes.all_shapes

    @classmethod
    def getHeight(cls, width, buttons):
        if cls.fitsScreen(width, buttons, 40):
            return 40
        return 20

    @classmethod
    def estimateWidth(cls, buttons, icon_size):
        nb_icons = len(buttons) + len(cls.available_colors) + len(cls.available_shapes)
        return icon_size * nb_icons + cls.margin * (nb_icons - 4) + 2 * cls.separation

    @classmethod
    def fitsScreen(cls, width, buttons, icon_size):
        return not width < cls.estimateWidth(buttons, icon_size)

    def __init__(self, screen, buttons: list, icon_size: int):
        self.screen = screen
        self.rect = pygame.Rect(0, 0, screen.get_width(), icon_size)
        self.icon_size = icon_size
        self.selected_color = self.available_colors[0]
        self.selected_shape = self.available_shapes[0]
        self.available_buttons = buttons

        x_offset = self.margin

        self.icons = []
        for _ in self.available_shapes:
            self.icons.append(Icon(x_offset, 0, self.icon_size))
            x_offset += self.icon_size + self.margin

        x_offset += self.separation
        for _ in self.available_colors:
            self.icons.append(Icon(x_offset, 0, self.icon_size))
            x_offset += self.icon_size

        x_offset += self.separation
        for _ in self.available_buttons:
            self.icons.append(Icon(x_offset, 0, self.icon_size))
            x_offset += self.icon_size + self.margin

    def icon_action(self, index):
        nbs = len(self.available_shapes)
        nbc = len(self.available_colors)
        if index < nbs:
            self.selected_shape = self.available_shapes[index]
        elif index < nbs + nbc:
            self.selected_color = self.available_colors[index - nbs]
        else:
            self.available_buttons[index - nbs - nbc].execute()

    def draw(self):
        # Draw the status bar background

        icon_counter = itertools.count(0)

        radius = 3

        for shape in self.available_shapes:
            color = SELECTED_COLOR if shape == self.selected_shape else GREY
            idx = next(icon_counter)
            icon = self.icons[idx]
            pygame.draw.rect(self.screen, color, icon.rect, border_radius=radius)
            if icon.shape is None:
                icon.shape = shape(
                    color=BLACK,
                    start=icon.rect.topleft,
                    end=icon.rect.bottomright,
                    isDummy=True,
                )
            icon.shape.draw(self.screen)

        for color in self.available_colors:
            idx = next(icon_counter)
            icon = self.icons[idx]
            if color == self.selected_color:
                pygame.draw.rect(
                    self.screen, SELECTED_COLOR, icon.rect, border_radius=radius
                )
            if not getattr(icon, "surface", None):
                icon.surface = shapes.make_color_shape(
                    color, icon.rect.width, icon.rect.height
                )
            self.screen.blit(icon.surface, icon.rect)

        for but in self.available_buttons:
            idx = next(icon_counter)
            icon = self.icons[idx]
            but.draw(self.screen, icon.rect)

    def handle_event(self, event):
        if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
            for idx, shape in enumerate(self.icons):
                # detect if shape is under cursor
                if shape.rect.collidepoint(event.pos):
                    self.icon_action(idx)
                    return True



import os
import io
import pygame
import subprocess

from .colors import GREY, SELECTED_COLOR
from .widgets import Button
from . import shapes

OUTPUT_FILENAME = os.environ.get("ANNOTATED", "/tmp/annotated.jpg")


class BigSmallBut(Button):
    def __init__(self, gui):
        super().__init__(gui, "󰧴")

    def execute(self):
        if shapes.WIDGET_SCALE == 1:
            shapes.WIDGET_SCALE = 2
            self.bg = SELECTED_COLOR
        else:
            shapes.WIDGET_SCALE = 1
            self.bg = GREY
        if self.gui.objects:
            self.gui.dirty_annotation = True


class SaveBut(Button):
    def __init__(self, gui):
        super().__init__(gui, "")

    def execute(self):
        pygame.image.save(self.gui.get_annotated_image(), OUTPUT_FILENAME)


class CopyBut(Button):
    def __init__(self, gui):
        super().__init__(gui, "")

    def execute(self):
        if os.environ.get("WAYLAND_DISPLAY"):
            proc = subprocess.Popen(
                ["wl-copy", "-t", "image/png"], stdin=subprocess.PIPE
            )
        else:
            proc = subprocess.Popen(
                ["xclip", "-selection", "clipboard", "-t", "image/png"],
                stdin=subprocess.PIPE,
            )
        buffer = io.BytesIO()
        pygame.image.save(self.gui.get_annotated_image(), buffer, "png")
        proc.communicate(input=buffer.getvalue())


class BackBut(Button):
    def __init__(self, gui):
        super().__init__(gui, "󰕌")

    def execute(self):
        if self.gui.objects:
            self.gui.objects[-1].instance_removed()
            self.gui.objects = self.gui.objects[:-1]
            self.gui.dirty_annotation = True


class ClearBut(Button):
    def __init__(self, gui):
        super().__init__(gui, "󰗩")

    def execute(self):
        for obj in self.gui.objects:
            obj.instance_removed()
            # Clear all the shapes
        self.gui.objects = []
        self.gui.dirty_annotation = True
